﻿using ImageProcessingGraph.Editor.Nodes.NodeAttributes;
using UnityEngine;
using Unity.Collections;
using Unity.Jobs;
using Unity.Burst;

namespace ImageProcessingGraph.Editor.Nodes.Types.Image.Utilities.ViewNode
{
    [NodeInfo("View Texture", "Utility/View Texture", false, null , editorType: typeof(ViewTextureNodeEditor))]
    public partial class ViewTextureNode : BaseImageNode
    {
        [NodeAttributes.Input("Texture")] public Texture2D texture;
        [NodeAttributes.Input("Image Data")] public ImageData imageData;

        public delegate void OnImageUpdated();
        public OnImageUpdated onImageUpdated;

        public Texture2D cachedRGB;
        public Texture2D cachedR;
        public Texture2D cachedG;
        public Texture2D cachedB;
        public Texture2D cachedA;

        public override void Process()
        {
            if (texture == null)
            {
                texture = imageData.ToTexture2D();
            }

            GenerateChannelTextures(texture);

            onImageUpdated?.Invoke();
        }

        public override void CleanUp()
        {
            texture = null;
            imageData = default;
        }

        private void GenerateChannelTextures(Texture2D source)
        {
            var pixels = source.GetPixels();
            int length = pixels.Length;

            NativeArray<Color> inputPixels = new NativeArray<Color>(pixels, Allocator.TempJob);

            NativeArray<Color> outputR = new NativeArray<Color>(length, Allocator.TempJob);
            NativeArray<Color> outputG = new NativeArray<Color>(length, Allocator.TempJob);
            NativeArray<Color> outputB = new NativeArray<Color>(length, Allocator.TempJob);
            NativeArray<Color> outputA = new NativeArray<Color>(length, Allocator.TempJob);

            var jobR = new ChannelJob(ChannelType.R, inputPixels, outputR);
            var jobG = new ChannelJob(ChannelType.G, inputPixels, outputG);
            var jobB = new ChannelJob(ChannelType.B, inputPixels, outputB);
            var jobA = new ChannelJob(ChannelType.A, inputPixels, outputA);

            JobHandle handleR = jobR.Schedule(length, 64);
            JobHandle handleG = jobG.Schedule(length, 64);
            JobHandle handleB = jobB.Schedule(length, 64);
            JobHandle handleA = jobA.Schedule(length, 64);

            var handles = new NativeArray<JobHandle>(4, Allocator.Temp);
            handles[0] = handleR;
            handles[1] = handleG;
            handles[2] = handleB;
            handles[3] = handleA;

            JobHandle.CompleteAll(handles);
            handles.Dispose(); 


            cachedRGB = source;
            cachedR = CreateTexture(source.width, source.height, outputR);
            cachedG = CreateTexture(source.width, source.height, outputG);
            cachedB = CreateTexture(source.width, source.height, outputB);
            cachedA = CreateTexture(source.width, source.height, outputA);

            inputPixels.Dispose();
            outputR.Dispose();
            outputG.Dispose();
            outputB.Dispose();
            outputA.Dispose();
        }

        private Texture2D CreateTexture(int width, int height, NativeArray<Color> colors)
        {
            Texture2D tex = new Texture2D(width, height, TextureFormat.RGBA32, false);
            tex.SetPixels(colors.ToArray());
            tex.Apply();
            return tex;
        }

        [BurstCompile]
        private struct ChannelJob : IJobParallelFor
        {
            public ChannelType channelType;
            [ReadOnly] public NativeArray<Color> input;
            public NativeArray<Color> output;

            public ChannelJob(ChannelType type, NativeArray<Color> input, NativeArray<Color> output)
            {
                this.channelType = type;
                this.input = input;
                this.output = output;
            }

            public void Execute(int index)
            {
                Color c = input[index];
                switch (channelType)
                {
                    case ChannelType.R:
                        output[index] = new Color(c.r, c.r, c.r, 1f);
                        break;
                    case ChannelType.G:
                        output[index] = new Color(c.g, c.g, c.g, 1f);
                        break;
                    case ChannelType.B:
                        output[index] = new Color(c.b, c.b, c.b, 1f);
                        break;
                    case ChannelType.A:
                        output[index] = new Color(c.a, c.a, c.a, 1f);
                        break;
                }
            }
        }

        private enum ChannelType
        {
            R, G, B, A
        }
    }
}
